#include <sys/types.h>
#include <sys/capability.h>

typedef char   sm_cmdline[256];
typedef uid_t  sm_uid;
typedef gid_t  sm_gid;
typedef sm_gid sm_gids[16];
typedef pid_t  sm_pid;
typedef cap_t  sm_caps;
typedef char   sm_path[256];


#include <stdint.h>

#define H1(s,i,x)   (x*65599u+(uint8_t)s[(i)<sizeof(s)-1?sizeof(s)-2-(i):sizeof(s)-1])
#define H4(s,i,x)   H1(s,i,H1(s,i+1,H1(s,i+2,H1(s,i+3,x))))
#define H16(s,i,x)  H4(s,i,H4(s,i+4,H4(s,i+8,H4(s,i+12,x))))
#define H64(s,i,x)  H16(s,i,H16(s,i+16,H16(s,i+32,H16(s,i+48,x))))
#define H256(s,i,x) H64(s,i,H64(s,i+64,H64(s,i+128,H64(s,i+192,x))))
#define HASH(s)     ((uint32_t)(H256(s,0,0)^(H256(s,0,0)>>16)))
#define str(s)  xstr(s)
#define xstr(s) #s


#define SM_ID_OFF(t)  ((sizeof(t) - 1) * CHAR_BIT)
#define SM_ID_BITS(t) (CHAR_BIT - 1)
#define SM_ID_MASK(t) ((1 << SM_ID_BITS(t)) - 1)
#define SM_ID_HASH(t) (HASH(str(t)) & SM_ID_MASK(t))

#define SM_DECLARE_ID_TYPE(u, t) \
	typedef u t; \
	static inline t t##_new(u x) { \
		return (SM_ID_HASH(t) << SM_ID_OFF(u)) | x; \
	} \
	static inline u t##_val(t x) { \
		if ((x >> SM_ID_OFF(u)) & SM_ID_MASK(u) != SM_ID_HASH(t)) { \
			return -1; \
		} \
		return x & ~(SM_ID_MASK(u) << SM_ID_OFF(u));
	} \

typedef uint32_t sm_id;




typedef int    sm_res_proto;
typedef char   sm_res_id[256];

struct sm_res {
	sm_res_proto r_proto;
	sm_res_id    r_id;
};

typedef struct sm_res sm_resources[16];


enum sm_state {
	SM_ST_STARTED,
	SM_ST_RUNNING,
	SM_ST_FINISHED,
	SM_ST_ERRORED,
};


typedef struct sm_job;

enum sm_dep_strength {
	SM_DEP_USE,
	SM_DEP_WANT,
	SM_DEP_NEED,
};

struct sm_dep {
	struct sm_job *d_target;
	enum sm_dep_strength d_before;
	enum sm_dep_strength d_during;
	enum sm_dep_strength d_after;
};

struct sm_job {
	sm_cmdline   j_cmdline;
	sm_uid       j_uid;
	sm_gids      j_gids;
	sm_caps      j_caps;
	sm_path      j_root;
	sm_resources j_resources;
	sm_env       j_env;

	enum sm_state success_state;
};

struct sm_job {
	
};
