#include <errno.h>

#include <unistd.h>
#include <sys/socket.h>
#include "libreap.h"

static inline int sendmsg_with_fd(int sock, const void *buf, size_t buflen, int flags, int fd)
{
	union {
		struct cmsghdr cmsghdr;
		char control[CMSG_SPACE(sizeof(int))];
	} cmsg = { 0 };
	struct msghdr msg = {
		.msg_iov = &(struct iovec) {
			.iov_base = (void *)buf,
			.iov_len = buflen,
		},
		.msg_iovlen = 1,
		.msg_control = &cmsg,
		.msg_controllen = sizeof(cmsg),
	};
	struct cmsghdr *cp = CMSG_FIRSTHDR(&msg);
	cp->cmsg_level = SOL_SOCKET;
	cp->cmsg_type = SCM_RIGHTS;
	cp->cmsg_len = CMSG_LEN(sizeof(int));
	*(int *)CMSG_DATA(cp) = fd;
	return sendmsg(sock, &msg, flags);
}

int sm_reaper_open(void)
{
	int fd = SM_REAPER_FD;

	/* sanity check */
	int type;
	socklen_t typelen = sizeof(type);
	if (getsockopt(fd, SOL_SOCKET, SO_TYPE, &type, &typelen) < 0)
		return -errno;
	if (type != SOCK_STREAM && type != SOCK_SEQPACKET)
		return -EBADF;

	return fd;
}

int sm_reaper_add(int fd, const struct sm_reaper_proc *proc)
{
	pid_t pid = proc->pid;
	if (proc->fd >= 0)
		return sendmsg_with_fd(fd, &pid, sizeof(pid), MSG_NOSIGNAL | MSG_EOR, proc->fd);
	else
		return send(fd, &pid, sizeof(pid), MSG_NOSIGNAL | MSG_EOR);
}

int sm_reaper_get(int fd, struct sm_reaped_proc *proc)
{
	return recv(fd, proc, sizeof(*proc), MSG_DONTWAIT);
}

void sm_reaper_close(int fd)
{
	close(fd);
}
