#include <unistd.h>
#include <sys/types.h>
#include <fcntl.h>
#include <errno.h>

#include <sys/epoll.h>
#include <sys/pidfd.h>
#include <sys/wait.h>
#include "libreap.h"
#include "reaper-internal.h"

int sm_reaper_open(void)
{
	int fd = epoll_create(1);
	if (fd < 0)
		return -errno;
	(void)fcntl(fd, F_SETFD, FD_CLOEXEC);
	(void)sm_reaper_set_child_reaper(1);
	return fd;
}

int sm_reaper_add(int fd, const struct sm_reaper_proc *proc)
{
	int pfd = proc->fd;
	if (pfd < 0)
		pfd = pidfd_open(proc->pid, 0);
	if (pfd < 0)
		return -errno;

	struct epoll_event ev = {
		.events = EPOLLIN,
		.data = { .fd = pfd },
	};
	if (epoll_ctl(fd, EPOLL_CTL_ADD, pfd, &ev) < 0) {
		(void)close(pfd);
		return -errno;
	}
	return 0;
}

int sm_reaper_get(int fd, struct sm_reaped_proc *proc)
{
	/* get pidfd */
	struct epoll_event ev;
	int r;
	while ((r = epoll_wait(fd, &ev, 1, 0)) < 0) {
		if (errno != EINTR)
			return -errno;
	}
	if (r == 0)
		return 0;

	/* get process status */
	siginfo_t info = { .si_pid = 0 };
	while (waitid(P_PIDFD, ev.data.fd, &info, WEXITED | WNOHANG) < 0) {
		if (errno != EINTR)
			return -errno;
	}
	if (info.si_pid == 0)
		return 0;

	/* it exited; remove from interest list */
	(void)epoll_ctl(fd, EPOLL_CTL_DEL, ev.data.fd, &ev);
	(void)close(ev.data.fd)

	proc->pid = info.si_pid;
	proc->status = info.si_code;
	proc->code = info.si_status;
	return 1;
}

void sm_reaper_close(int fd)
{
	(void)sm_reaper_set_child_reaper(0);
	(void)close(fd);
}
