#include <stdatomic.h>
#include <unistd.h>
#include <sys/types.h>
#include <fcntl.h>
#include <errno.h>

#include <signal.h>
#include <sys/wait.h>
#include "libreap.h"
#include "reaper-internal.h"

int sm_reaper_open(void)
{
	sigset_t sigs;
	sigemptyset(&sigs);
	sigaddset(&sigs, SIGCHLD);

	int fd = signalfd(-1, &sigs, 0);
	if (fd < 0)
		return -errno;

	(void)fcntl(fd, F_SETFD, FD_CLOEXEC);
	int options = fcntl(fd, F_GETFL, NULL);
	if (options >= 0)
		(void)fcntl(fd, F_SETFL, options | O_NONBLOCK);
	(void)sm_reaper_set_child_reaper(1);
	return fd;
}

int sm_reaper_add(int fd, const struct sm_reaper_proc *proc)
{
	/* not needed */
	return 0;
}

int sm_reaper_get(int fd, struct sm_reaped_proc *proc)
{
	/* get signal info */
	struct signalfd_siginfo info;
	for (;;) {
		int r;
		while ((r = read(fd, &info, sizeof(info))) < 0) {
			if (errno != EINTR)
				return -errno;
		}
		if (r == 0)
			return 0;
		if (info.ssi_signo != SIGCHLD || info.ssi_pid == (pid_t)0)
			continue;

		proc->pid = info.ssi_pid;
		int stat = info.ssi_status;
		if (WIFEXITED(stat)) {
			proc->status = CLD_EXITED;
			proc->code = WEXITSTATUS(stat);
		} else if (WIFSIGNALED(stat)) {
			proc->status = WCOREDUMP(stat) ? CLD_DUMPED : CLD_KILLED;
			proc->code = WTERMSIG(stat);
		} else
			continue;
		return 1;
	}
}

void sm_reaper_close(int fd)
{
	(void)sm_reaper_set_child_reaper(0);
	(void)close(fd);
}
